import time

import gevent
import gevent.socket
import gevent.queue
import gevent.wsgi
from aspen.network_engines import CooperativeEngine
from aspen.sockets import packet
from aspen.sockets.loop import Die


class GeventBuffer(gevent.queue.Queue):
    """Model a buffer of items.
  
    There are two of these for each Socket, one for incoming message payloads
    and one for outgoing message objects.

    Here's what the flow looks like: 

        wire => [msg, msg, msg, msg, msg, msg, msg, msg] => resource
        wire <= [msg, msg, msg, msg, msg, msg, msg, msg] <= resource

    """

    def __init__(self, name, socket=None):
        """Takes a string and maybe a socket.

        If given a socket, we will try to play nice with its loop.

        """
        gevent.queue.Queue.__init__(self)
        self._socket = socket
        self._name = name


    # flush
    # =====
    # Used for outgoing buffer.

    def flush(self):
        """Return an iterable of bytestrings or None.
        """
        if not self.empty():
            return self.__flusher()
        return None

    def __flusher(self):
        """Yield strings.

        We unload bytestrings as fast as we can until we run out of time or
        bytestrings. On my MacBook Pro I am seeing between 500 and 1000
        messages dumped in 2ms--without any WSGI/HTTP/TCP overhead. We always
        yield at least one bytestring to avoid deadlock.

        This generator is instantiated in self.flush.

        """
        if not self.empty():
            yield packet.frame(self.get())
        timeout = time.time() + (0.007) # We have 7ms to dump bytestrings. Go!
        while not self.empty() and time.time() < timeout:
            yield packet.frame(self.get())


    # next 
    # ====
    # Used for incoming buffer.

    def next(self):
        """Return the next item from the queue.
        
        The first time this is called, we lazily instantiate the generator at
        self._blocked. Subsequent calls are directed directly to that
        generator's next method.

        """
        self._blocked = self._blocked()
        self.next = self._next
        return self.next()

    def _next(self):
        try:
            return self._blocked.next()
        except StopIteration:
            # When the _blocked generator discovers Die and breaks, the 
            # effect is a StopIteration here. It's a bug if this happens
            # other than when we are disconnecting the socket.
            assert self._socket is not None
            assert self._socket.loop.please_stop

    def _blocked(self):
        """Yield items from self forever.

        This generator is lazily instantiated in self.next. It is designed to
        cooperate with ThreadedLoop.

        """
        if self._socket is None:    # We're on a Channel.
            while 1:
                yield self.get()
        else:                       # We're on a Socket.
            while not self._socket.loop.please_stop:
                out = self.get()
                if out is Die:
                    break # will result in a StopIteration
                yield out


class GeventLoop(object):
   
    def __init__(self, socket):
        self.socket = socket
        self.please_stop = False
        self.greenthread = None

    def __call__(self):
        while not self.please_stop:
            self.socket.tick()

    def start(self):
        self.greenthread = gevent.spawn(self)

    def stop(self):
        self.please_stop = True
        self.socket.incoming.put(Die)
        self.greenthread.wait()


class Engine(CooperativeEngine):

    wsgi_server = None # a WSGI server, per gevent

    def bind(self):
        self.gevent_server = gevent.wsgi.WSGIServer( listener=self.website.network_address
                                                   , application=self.website
                                                   , log=None
                                                    )

    def sleep(self, seconds):
        gevent.sleep(seconds)

    def start(self):
        self.gevent_server.serve_forever()

    def start_restarter(self, check_all):
        def loop():
            while True:
                check_all()
                self.sleep(0.5)
        gevent.spawn(loop)

    Buffer = GeventBuffer
    Loop = GeventLoop




import time
import threading

from aspen._cherrypy.wsgiserver import CherryPyWSGIServer
from aspen.network_engines import ThreadedEngine


class Engine(ThreadedEngine):

    cp_server = None # a CherryPyWSGIServer instance

    def bind(self):
        self.cp_server = CherryPyWSGIServer( self.website.network_address
                                           , self.website
                                            )

        # Work around a Jython bug.
        # =========================
        # http://bugs.jython.org/issue1848
        # http://stackoverflow.com/questions/1103487/
        try:                        # >= 2.6
            import platform
            if platform.python_implementation() == 'Jython':
                self.cp_server.nodelay = False
        except AttributeError:      # < 2.6
            import sys
            if sys.platform.startswith('java'):
                self.cp_server.nodelay = False

    def start(self):
        self.cp_server.start()

    def stop(self):
        self.cp_server.stop()

    def start_restarter(self, check_all):

        def loop():
            while True:
                try:
                    check_all()
                except SystemExit:
                    self.cp_server.interrupt = SystemExit
                time.sleep(0.5)

        checker = threading.Thread(target=loop)
        checker.daemon = True
        checker.start()



from Cookie import CookieError, SimpleCookie

from aspen.http.mapping import CaseInsensitiveMapping


class BaseHeaders(CaseInsensitiveMapping):
    """Represent the headers in an HTTP Request or Response message.
    """

    def __init__(self, raw):
        """Takes headers as a string.
        """
        def genheaders():
            for line in raw.splitlines():
                k, v = line.split(':', 1)
                yield k.strip(), v.strip()
        CaseInsensitiveMapping.__init__(self, genheaders)
      

        # Cookie
        # ======

        self.cookie = SimpleCookie()
        try:
            self.cookie.load(self.get('Cookie', ''))
        except CookieError:
            pass # XXX really?


    def raw(self):
        """Return the headers as a string, formatted for an HTTP message.
        """
        out = []
        for header, values in self.iteritems():
            for value in values:
                out.append('%s: %s' % (header, value))
        return '\r\n'.join(out)
    raw = property(raw)



import pants
from aspen.network_engines import CooperativeEngine
from pants.contrib.http import HTTPServer
from pants.contrib.wsgi import WSGIConnector


class Engine(CooperativeEngine):

    def bind(self):
        connector = WSGIConnector(self.website)
        self.server = HTTPServer(connector)
        self.server.listen( host=self.website.network_address[0]
                          , port=self.website.network_address[1]
                           )

    def start(self):
        pants.engine.start()

    def stop(self):
        pants.engine.stop()

    def start_restarter(self, check_all):
        pants.cycle(check_all, 0.5)



import time
import threading

import cheroot.wsgi
from aspen.network_engines import ThreadedEngine


class Engine(ThreadedEngine):

    cheroot_server = None

    def bind(self):
        name = "Aspen! Cheroot!"
        self.cheroot_server = cheroot.wsgi.WSGIServer( self.website.network_address
                                                     , server_name=name
                                                     , wsgi_app=self.website
                                                      )

    def start(self):
        self.cheroot_server.start()

    def stop(self):
        self.cheroot_server.stop()

    def start_restarter(self, check_all):

        def loop():
            while True:
                try:
                    check_all()
                except SystemExit:
                    self.cheroot_server.interrupt = SystemExit
                time.sleep(0.5)

        checker = threading.Thread(target=loop)
        checker.daemon = True
        checker.start()



"""Oh, wow.
"""
import time

import twisted.internet
import twisted.internet.task
import twisted.web.server 
import twisted.web.wsgi 
from aspen.network_engines import ThreadedEngine


class Engine(ThreadedEngine):
    """For v1 we're just running a threaded server. Maybe some day ...
    """

    twisted_site = None

    def bind(self):
        pool = twisted.internet.reactor.getThreadPool()
        resource = twisted.web.wsgi.WSGIResource( twisted.internet.reactor
                                                , pool
                                                , self.website 
                                                 )
        self.twisted_site = twisted.web.server.Site(resource)

    def sleep(self, seconds):
        time.sleep(seconds) # We're threaded.

    def start(self):
        twisted.internet.reactor.listenTCP( port=self.website.network_address[1]
                                          , factory=self.twisted_site
                                          , interface=self.website.network_address[0]
                                           )
        twisted.internet.reactor.run()

    def start_restarter(self, check_all):
        self.restarter = twisted.internet.task.LoopingCall(check_all)
        deferred = self.restarter.start(0.5)

        # Set up twisted error handling.
        # ==============================
        # Twisted captures all exceptions, including SystemExit. Glyph actually
        # just told me a story (PyCon 2012 sprints) that when he was a kid his
        # father told him that programs should be robust in the face of error
        # conditions. "If you shoot it in the leg, it should crawl towards you.
        # Then when you crush its torso, it should drag itself towards you.
        # And when you shoot it in the head, it should reach out for you with 
        # its dying breath." So he captures SystemExit, and now, due to a very 
        # wonderful time together, I can respect that. And that's the way it 
        # works, motherfuckers. Get to it!
        # 
        # Other memorable quotes:
        # 
        # "Twisted started as a MUD written in Java that used three or four 
        #  threads per connection."
        # "That was the big moment: if NetHack can do it, so can I!"
        # "No, this wasn't ncurses, this was a full-on Swing app."
        # "1999."
        # "We had this cockroach autonomous character that would go around and
        #  infest your character's possessions, and everyone ended up with, 
        #  like, five of these things in their stuff."
        # "I just want there to be a single Python networking library with a
        #  consistent API."
        # "Oh! That's your vim. I thought for a second that Aspen was the most 
        #  advanced web server in the world."
        # "I started using vim when I was six."

        def error(failure):
            failure.trap(SystemExit)
            if failure.type is SystemExit and self.restarter.running:
                self.restarter.stop()

        deferred.addErrback(error)

    def stop_restarter(self):
        self.restarter.stop()



import sys

try:                # Python >= 2.6
    from collections import Callable
    def is_callable(obj):
        return isinstance(obj, Callable)
except ImportError: # Python < 2.6
    from operator import isCallable as is_callable

# imports of convenience
from aspen.http.response import Response
from aspen import json_ as json
from aspen.logging import log, log_dammit

# Shut up, PyFlakes. I know I'm addicted to you.
Response, json, is_callable, log, log_dammit


__version__ = "~~VERSION~~"
WINDOWS = sys.platform[:3] == 'win'
NETWORK_ENGINES = ['cheroot', 'cherrypy', 'diesel', 'eventlet', 'gevent',
                   'pants', 'rocket', 'tornado', 'twisted']
RENDERERS = ['jinja2', 'pystache', 'tornado']



import codecs
import re


# Register a 'repr' error strategy.
# =================================
# Sometimes we want to echo bytestrings back to a user, and we don't know what 
# encoding will work. This error strategy replaces non-decodable bytes with 
# their Python representation, so that they are human-visible.
# 
# See also:
#   - https://github.com/dcrosta/mongo/commit/e1ac732
#   - http://www.crummy.com/software/BeautifulSoup/bs4/doc/#unicode-dammit

def replace_with_repr(unicode_error):
    offender = unicode_error.object[unicode_error.start:unicode_error.end]
    return (unicode(repr(offender).strip("'").strip('"')), unicode_error.end)

codecs.register_error('repr', replace_with_repr)


def unicode_dammit(s, encoding="UTF-8"):
    """Given a bytestring, return a unicode decoded with `encoding`.

    Any bytes not decodable with UTF-8 will be replaced with their Python
    representation, so you'll get something like u"foo\\xefbar".

    """
    if not isinstance(s, str):
        raise TypeError("I got %s, but I want <type 'str'>." % s.__class__)
    errors = 'repr'
    return s.decode(encoding, errors)


def ascii_dammit(s):
    """Given a bytestring, return a bytestring.

    The returned bytestring will have any non-ASCII bytes replaced with
    their Python representation, so it will be pure ASCII.

    """
    return unicode_dammit(s, encoding="ASCII").encode("ASCII")


# typecheck

def typecheck(*checks):
    """Assert that arguments are of a certain type.

    Checks is a flattened sequence of objects and target types, like this:

        ( {'foo': 2}, dict
        , [1,2,3], list
        , 4, int
        , True, bool
        , 'foo', (basestring, None)
         )

    The target type can be a single type or a tuple of types. None is
    special-cased (you can specify None and it will be interpreted as
    type(None)).

    >>> typecheck()
    >>> typecheck('foo')
    Traceback (most recent call last):
        ...
    AssertionError: typecheck takes an even number of arguments.
    >>> typecheck({'foo': 2}, dict)
    >>> typecheck([1,2,3], list)
    >>> typecheck(4, int)
    >>> typecheck(True, bool)
    >>> typecheck('foo', (str, None))
    >>> typecheck(None, None)
    >>> typecheck(None, type(None))
    >>> typecheck('foo', unicode)
    Traceback (most recent call last):
        ...
    TypeError: 'foo' is of type str, but unicode was expected.
    >>> typecheck('foo', (basestring, None))
    Traceback (most recent call last):
        ...
    TypeError: 'foo' is of type str, not one of: basestring, NoneType.

    """
    assert type(checks) is tuple, checks
    assert len(checks) % 2 == 0, "typecheck takes an even number of arguments."

    def nice(t):
        return re.findall("<type '(.+)'>", str(t))[0]

    checks = list(checks)
    checks.reverse()

    while checks:
        obj = checks.pop()
        expected = checks.pop()
        actual = type(obj)

        if isinstance(expected, tuple):
            expected = list(expected)
        elif not isinstance(expected, list):
            expected = [expected]

        for i, t in enumerate(expected):
            if t is None:
                expected[i] = type(t)
        
        if actual not in expected: 
            msg = "%s is of type %s, " % (repr(obj), nice(actual))
            if len(expected) > 1:
                niced = [nice(t) for t in expected]
                msg += ("not one of: %s." % ', '.join(niced))
            else:
                msg += "but %s was expected." % nice(expected[0])
            raise TypeError(msg)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
